# -*- coding: utf-8 -*-
"""
Methods to help validate a given JSON document against the Swagger Spec.
"""
import os

import swagger_spec_validator
from jsonschema.exceptions import ValidationError

from .exceptions import wrap_exception

API_DOCS_FILENAME = 'api_docs.json'


@wrap_exception(ValidationError)
def validate_swagger_schema(schema_dir, resource_listing):
    """Validate the structure of Swagger schemas against the spec.

    **Valid only for Swagger v1.2 spec**

    Note: It is possible that resource_listing is not present in
    the schema_dir. The path is passed in the call so that ssv
    can fetch the api-declaration files from the path.

    :param resource_listing: Swagger Spec v1.2 resource listing
    :type resource_listing: dict
    :param schema_dir: A path to Swagger spec directory
    :type schema_dir: string
    :raises: :py:class:`swagger_spec_validator.SwaggerValidationError`
    """
    schema_filepath = os.path.join(schema_dir, API_DOCS_FILENAME)
    swagger_spec_validator.validator12.validate_spec(
        resource_listing,
        "file://{0}".format(os.path.abspath(schema_filepath)))



# -*- coding: utf-8 -*-
"""
This model contains the main factory of renderers to use while dealing with Swagger 2.0 endpoints.
"""
from functools import partial

from bravado_core.exception import MatchingResponseNotFound
from bravado_core.exception import SwaggerMappingError
from bravado_core.marshal import marshal_schema_object
from bravado_core.response import get_response_spec
from pyramid.renderers import JSON


class PyramidSwaggerRendererFactory(object):
    def __init__(self, renderer_factory=JSON()):
        self.renderer_factory = renderer_factory

    def _marshal_object(self, request, response_object):
        # operation attribute is injected by validator_tween in case the endpoint is served by Swagger 2.0 specs
        operation = getattr(request, 'operation')

        if not operation:
            # If the request is not served by Swagger2.0 endpoint _marshal_object is NO_OP
            return response_object

        try:
            response_spec = get_response_spec(
                status_code=request.response.status_code,
                op=request.operation,
            )
            return marshal_schema_object(
                swagger_spec=request.registry.settings['pyramid_swagger.schema20'],
                schema_object_spec=response_spec['schema'],
                value=response_object,
            )
        except MatchingResponseNotFound:
            # Response specs not found
            return response_object
        except SwaggerMappingError:
            # marshaling process failed
            return response_object

    def _render(self, external_renderer, value, system):
        value = self._marshal_object(system['request'], value)
        return external_renderer(value, system)

    def __call__(self, info):
        return partial(self._render, self.renderer_factory(info))



# -*- coding: utf-8 -*-
"""
The core model we use to represent the entire ingested swagger schema for this
service.
"""
import re
from collections import namedtuple


PyramidEndpoint = namedtuple(
    'PyramidEndpoint',
    'path route_name view renderer')


class PathNotMatchedError(Exception):
    """Raised when a SwaggerSchema object is given a request it cannot match
    against its stored schema."""
    pass


class SwaggerSchema(object):
    """
    This object contains data structures representing your Swagger schema
    and exposes methods for efficiently finding the relevant schemas for a
    Pyramid request.

    :param pyramid_endpoints: a list of :class:`PyramidEndpoint` which define
        the pyramid endpoints to create for serving the api docs
    :param resource_validators: a list of resolvers, one per Swagger resource
    :type resource_validators: list of mappings from :class:`RequestMatcher`
        to :class:`ValidatorMap`
    for every operation in the api specification.
    """

    def __init__(self, pyramid_endpoints, resource_validators):
        self.pyramid_endpoints = pyramid_endpoints
        self.resource_validators = resource_validators

    def validators_for_request(self, request, **kwargs):
        """Takes a request and returns a validator mapping for the request.

        :param request: A Pyramid request to fetch schemas for
        :type request: :class:`pyramid.request.Request`
        :returns: a :class:`pyramid_swagger.load_schema.ValidatorMap` which can
            be used to validate `request`
        """
        for resource_validator in self.resource_validators:
            for matcher, validator_map in resource_validator.items():
                if matcher.matches(request):
                    return validator_map

        raise PathNotMatchedError(
            'Could not find the relevant path ({0}) in the Swagger schema. '
            'Perhaps you forgot to add it?'.format(request.path_info)
        )

    def get_api_doc_endpoints(self):
        return self.pyramid_endpoints


def partial_path_match(path1, path2, kwarg_re=r'\{.*\}'):
    """Validates if path1 and path2 matches, ignoring any kwargs in the string.

    We need this to ensure we can match Swagger patterns like:
        /foo/{id}
    against the observed pyramid path
        /foo/1

    :param path1: path of a url
    :type path1: string
    :param path2: path of a url
    :type path2: string
    :param kwarg_re: regex pattern to identify kwargs
    :type kwarg_re: regex string
    :returns: boolean
    """
    split_p1 = path1.split('/')
    split_p2 = path2.split('/')
    pat = re.compile(kwarg_re)
    if len(split_p1) != len(split_p2):
        return False
    for partial_p1, partial_p2 in zip(split_p1, split_p2):
        if pat.match(partial_p1) or pat.match(partial_p2):
            continue
        if not partial_p1 == partial_p2:
            return False
    return True



# -*- coding: utf-8 -*-
import io
import os

from setuptools import find_packages
from setuptools import setup


base_dir = os.path.dirname(__file__)
about = {}
with open(os.path.join(base_dir, "pyramid_swagger", "__about__.py")) as f:
    exec(f.read(), about)

with io.open(os.path.join(base_dir, "README.rst"), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],

    description=about['__summary__'],
    long_description=long_description,
    license=about['__license__'],
    url=about["__uri__"],

    author=about['__author__'],
    author_email=about['__email__'],

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='pyramid swagger validation',
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    include_package_data=True,
    install_requires=[
        'bravado-core >= 4.8.4',
        'jsonschema',
        'pyramid',
        'simplejson',
    ],
)



# -*- coding: utf-8 -*-
"""
Module to load swagger specs and build efficient data structures for querying
them during request validation.
"""
from __future__ import unicode_literals

from collections import namedtuple

import simplejson
from jsonschema import _validators
from jsonschema import RefResolver
from jsonschema import validators
from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft3Validator
from jsonschema.validators import Draft4Validator

from pyramid_swagger.model import partial_path_match


EXTENDED_TYPES = {
    'number': (float,),
    'float': (float,),
    'int': (int,),
}


def build_param_schema(schema, param_type):
    """Turn a swagger endpoint schema into an equivalent one to validate our
    request.

    As an example, this would take this swagger schema:
        {
            "paramType": "query",
            "name": "query",
            "description": "Location to query",
            "type": "string",
            "required": true
        }
    To this jsonschema:
        {
            "type": "object",
            "additionalProperties": "False",
            "properties:": {
                "description": "Location to query",
                "type": "string",
                "required": true
            }
        }
    Which we can then validate against a JSON object we construct from the
    pyramid request.
    """
    properties = filter_params_by_type(schema, param_type)
    if not properties:
        return

    # Generate a jsonschema that describes the set of all query parameters. We
    # can then validate this against dict(request.params).
    return {
        'type': 'object',
        'properties': dict((p['name'], p) for p in properties),
        # Allow extra headers. Most HTTP requests will have headers which
        # are outside the scope of the spec (like `Host`, or `User-Agent`)
        'additionalProperties': param_type == 'header',
    }


def filter_params_by_type(schema, param_type):
    return [s for s in schema['parameters'] if s['paramType'] == param_type]


def extract_body_schema(schema):
    """Return the body parameter schema from an operation schema."""
    matching_body_schemas = filter_params_by_type(schema, 'body')
    # There can be only one body param
    return matching_body_schemas[0] if matching_body_schemas else None


def ignore(_validator, *args):
    """A validator which performs no validation. Used to `ignore` some schema
    fields during validation.
    """
    return


def build_swagger_type_validator(models):
    def swagger_type_validator(validator, ref, instance, schema):
        func = _validators.ref if ref in models else _validators.type_draft4
        return func(validator, ref, instance, schema)

    return swagger_type_validator


def type_validator(validator, types, instance, schema):
    """Swagger 1.2 supports parameters of 'type': 'File'. Skip validation of
    the 'type' field in this case.
    """
    if schema.get('type') == 'File':
        return []
    return _validators.type_draft3(validator, types, instance, schema)


def required_validator(validator, req, instance, schema):
    """Swagger 1.2 expects `required` to be a bool in the Parameter object, but
    a list of properties in a Model object.
    """
    if schema.get('paramType'):
        if req is True and not instance:
            return [ValidationError("%s is required" % schema['name'])]
        return []
    return _validators.required_draft4(validator, req, instance, schema)


def get_body_validator(models):
    """Returns a validator for the request body, based on a
    :class:`jsonschema.validators.Draft4Validator`, with extra validations
    added for swaggers extensions to jsonschema.

    :param models: a mapping of reference to models
    :returns: a :class:`jsonschema.validators.Validator` which can validate
        the request body.
    """
    return validators.extend(
        Draft4Validator,
        {
            'paramType': ignore,
            'name': ignore,
            'type': build_swagger_type_validator(models),
            'required': required_validator,
        }
    )


Swagger12ParamValidator = validators.extend(
    Draft3Validator,
    {
        'paramType': ignore,
        'name': ignore,
        'type': type_validator,
    }
)


class ValidatorMap(
    namedtuple('_VMap', 'query path form headers body response')
):
    """
    A data object with validators for each part of the request and response
    objects. Each field is a :class:`SchemaValidator`.
    """
    __slots__ = ()

    @classmethod
    def from_operation(cls, operation, models, resolver):
        args = []
        for schema, validator in [
            (build_param_schema(operation, 'query'), Swagger12ParamValidator),
            (build_param_schema(operation, 'path'), Swagger12ParamValidator),
            (build_param_schema(operation, 'form'), Swagger12ParamValidator),
            (build_param_schema(operation, 'header'), Swagger12ParamValidator),
            (extract_body_schema(operation), get_body_validator(models)),
            (extract_response_body_schema(operation, models),
                Draft4Validator),
        ]:
            args.append(SchemaValidator.from_schema(
                schema,
                resolver,
                validator))

        return cls(*args)


class SchemaValidator(object):
    """A Validator used by :mod:`pyramid_swagger.tween` to validate a
    field from the request or response.

    :param schema: a :class:`dict` jsonschema that was used by the
        validator
    :param valdiator: a Validator which a func:`validate` method
        for validating a field from a request or response. This
        will often be a :class:`jsonschema.validator.Validator`.
    """

    def __init__(self, schema, validator):
        self.schema = schema
        self.validator = validator

    @classmethod
    def from_schema(cls, schema, resolver, validator_class):
        return cls(
            schema,
            validator_class(schema, resolver=resolver, types=EXTENDED_TYPES))

    def validate(self, values):
        """Validate a :class:`dict` of values. If `self.schema` is falsy this
        is a noop.
        """
        if not self.schema:
            return
        self.validator.validate(values)


def build_request_to_validator_map(schema, resolver):
    """Build a mapping from :class:`RequestMatcher` to :class:`ValidatorMap`
    for each operation in the API spec. This mapping may be used to retrieve
    the appropriate validators for a request.
    """
    schema_models = schema.get('models', {})
    return dict(
        (
            RequestMatcher(api['path'], operation['method']),
            ValidatorMap.from_operation(operation, schema_models, resolver)
        )
        for api in schema['apis']
        for operation in api['operations']
    )


class RequestMatcher(object):
    """Match a :class:`pyramid.request.Request` to a swagger Operation"""

    def __init__(self, path, method):
        self.path = path
        self.method = method

    def matches(self, request):
        """
        :param request: a :class:`pyramid.request.Request`
        :returns: True if this matcher matches the request, False otherwise
        """
        return (
            partial_path_match(request.path_info, self.path) and
            request.method == self.method
        )


def extract_response_body_schema(operation, schema_models):
    if operation['type'] in schema_models:
        return extract_validatable_type(operation['type'], schema_models)

    acceptable_fields = (
        'type', '$ref', 'format', 'defaultValue', 'enum', 'minimum',
        'maximum', 'items', 'uniqueItems'
    )

    return dict(
        (field, operation[field])
        for field in acceptable_fields
        if field in operation
    )


def extract_validatable_type(type_name, models):
    """Returns a jsonschema-compatible typename from the Swagger type.

    This is necessary because for our Swagger specs to be compatible with
    swagger-ui, they must not use a $ref to internal models.

    :returns: A key-value that jsonschema can validate. Key will be either
        'type' or '$ref' as is approriate.
    :rtype: dict
    """
    if type_name in models:
        return {'$ref': type_name}
    else:
        return {'type': type_name}


def load_schema(schema_path):
    """Prepare the api specification for request and response validation.

    :returns: a mapping from :class:`RequestMatcher` to :class:`ValidatorMap`
        for every operation in the api specification.
    :rtype: dict
    """
    with open(schema_path, 'r') as schema_file:
        schema = simplejson.load(schema_file)
    resolver = RefResolver('', '', schema.get('models', {}))
    return build_request_to_validator_map(schema, resolver)



# -*- coding: utf-8 -*-
import sys

import six
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.httpexceptions import HTTPNotFound


class RequestValidationError(HTTPBadRequest):
    def __init__(self, *args, **kwargs):
        self.child = kwargs.pop('child', None)
        super(RequestValidationError, self).__init__(*args, **kwargs)


class PathNotFoundError(HTTPNotFound):
    def __init__(self, *args, **kwargs):
        self.child = kwargs.pop('child', None)
        super(PathNotFoundError, self).__init__(*args, **kwargs)


class ResponseValidationError(HTTPInternalServerError):
    def __init__(self, *args, **kwargs):
        self.child = kwargs.pop('child', None)
        super(ResponseValidationError, self).__init__(*args, **kwargs)


def wrap_exception(exception_class):
    def generic_exception(method):
        def wrapper(*args, **kwargs):
            try:
                method(*args, **kwargs)
            except Exception as e:
                six.reraise(
                    exception_class,
                    exception_class(str(e)),
                    sys.exc_info()[2])
        return wrapper
    return generic_exception
